# coding: utf8
""" This module gathers scores plotting functions. """

from __future__ import print_function
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator
from matplotlib import dates
import matplotlib.lines as mlines
from matplotlib import cm
import evaltools as evt
from datetime import datetime, date, timedelta, time
from scipy.stats import gaussian_kde, ttest_ind, mannwhitneyu
from scipy.stats import scoreatpercentile
from functools import reduce

# matplotlib figure size (width, height in inches)
figsize = (8, 5)


def _checkPeriod(objects):
    """ Check if Evaluator objects are defined on the same period. """
    starts = np.array([obj.startDate for obj in objects])
    ends = np.array([obj.endDate for obj in objects])
    if not (starts == starts[0]).all() or not (ends == ends[0]).all():
        raise evt.EvaltoolsError(
            "All Evaluator objects must be defined on the same period !!!")


def _checkStationList(objects):
    """ Check if Evaluator objects use the same station list. """
    staUnion = reduce(np.union1d,
                      [np.array(obj.stations.index) for obj in objects])
    staInter = reduce(np.intersect1d,
                      [np.array(obj.stations.index) for obj in objects])
    if staUnion.shape[0] != staInter.shape[0]:
        print("Warning: All Evaluator objects do not use " +
              "the same station list !!!")


def _defaultParams(objects, labels, colors, linestyles, markers,
                   default_marker=''):
    """ Define default parameters in several plotting functions. """
    # check if all objects have the same seriesType
    types = np.array([obj.seriesType for obj in objects])
    if (types == 'hourly').all():
        seriesType = 'hourly'
    elif (types == 'daily').all():
        seriesType = 'daily'
    else:
        raise evt.EvaltoolsError(
            "All objects must have the same seriesType attribute !!!")

    # define objects labels in case of None
    if labels is None:
        labels = []
        for obj in objects:
            labels.append(obj.model)

    # define objects colors in case of None
    if colors is None:
        colors = []
        for obj in objects:
            colors.append(obj.color)

    # define linestyles in case of None
    if linestyles is None:
        linestyles = ['-']*len(objects)

    # define markers in case of None
    if markers is None:
        markers = [default_marker]*len(objects)

    # choose nb of columns in legend
    if len(objects) > 4:
        legend_ncol = 2
    else:
        legend_ncol = 1

    # paint in black y=0 and x=0
    plt.axhline(y=0, color='k')
    plt.axvline(x=0, color='k')

    return labels, colors, linestyles, markers, legend_ncol, seriesType


def _save_figure(outputFile, file_formats, **kwargs):
    """ Save the figure in different formats or show it if outputFile is None.

    Parameters
    ----------
    outputFile : None or str
        Path of the file where to save the plots (without extension).
    file_formats : list of str
        List of file extensions among 'svgz', 'ps', 'emf', 'rgba', 'raw',
        'pdf', 'svg', 'eps' or 'png'.
    **kwargs :
        Keys of these arguments must be contained in the name of the output
        file and will be replaced by the corresponding values.

    """
    if outputFile is None:
        plt.show()
    else:
        for key in kwargs.keys():
            if "{" + key + "}" not in outputFile:
                raise evt.EvaltoolsError(
                    "'{" + key + "}' must be " +
                    "contained in the name of the output file !!!")
        for ext in file_formats:
            plt.savefig((outputFile+".{ext}").format(ext=ext, **kwargs))


def _is_outlier(points, thresh=3.5):
    """ Look for outliers.

    Returns a boolean array with True if points are outliers and False
    otherwise.

    Parameters
    ----------
    points : An numobservations by numdimensions array of observations
    thresh : The modified z-score to use as a threshold. Observations with
        a modified z-score (based on the median absolute deviation) greater
        than this value will be classified as outliers.

    Returns
    -------
        A numobservations-length boolean array.

    References
    ----------
        Boris Iglewicz and David Hoaglin (1993), "Volume 16: How to Detect and
        Handle Outliers", The ASQC Basic References in Quality Control:
        Statistical Techniques, Edward F. Mykytka, Ph.D., Editor.

    """
    if len(points.shape) == 1:
        points = points[:, None]
    median = np.nanmedian(points, axis=0)
    diff = np.nansum((points - median)**2, axis=-1)
    diff = np.sqrt(diff)
    med_abs_deviation = np.nanmedian(diff)

    modified_z_score = 0.6745*diff/med_abs_deviation
    # print(modified_z_score)

    return modified_z_score > thresh


def _black_axes(ax):
    """ Paint in black y=0 and x=0 """
    if ax.axes.get_xlim()[0] < 0 and ax.axes.get_ylim()[1] > 0:
        plt.axvline(x=0, color='k')
    if ax.axes.get_ylim()[0] < 0 and ax.axes.get_ylim()[1] > 0:
        plt.axhline(y=0, color='k')


def _set_minor_ticks(nb_x, nb_y):
    """ Define number of minor ticks between each major one. """
    ax = plt.gca()
    minorLocator = AutoMinorLocator(n=nb_x)
    ax.xaxis.set_minor_locator(minorLocator)
    minorLocator = AutoMinorLocator(n=nb_y)
    ax.yaxis.set_minor_locator(minorLocator)


class _Curves(object):
    """ Class designed to collect and work on the different curves in a chart.

    Parameters
    ----------
    x : 1D array-like
        Values for the x axis.

    """
    def __init__(self, x, ax=None):
        # abscissa
        self.x = x
        # curves
        self.style = {}
        self.curves = pd.DataFrame(index=self.x)
        self.curves.index.name = 'X'
        if ax is None:
            plt.close('all')
            self.fig = plt.figure(figsize=figsize)
            self.ax = self.fig.add_subplot(1, 1, 1)
        else:
            self.ax = ax
            self.fig = ax.figure
        # draw grid
        self.ax.grid(which='both', ls='-', c='#B0B0B0')
        # try to control number of yticks
        self.ax.locator_params(axis='y', nbins=11, min_n_ticks=10)

    def add_curve(self, key, y, style={}):
        """ Add a curve. """
        self.curves[key] = y
        self.style[key] = style

    def plot(self, black_axes=False, outlier_thresh=None, legend=True):
        """ Plot curves.

        Parameters
        ----------
        black_axes : bool
            If true, y=0 and x=0 lines are painted in black.
        outlier_thresh : scalar or None
            If not None, it correspond to the threshold used in
            evaltools.plotting._is_outlier to determine if a model is an
            outlier. If outliers are detected, y boudaries do not take
            them into account.

        """
        for key in self.curves.columns:
            plt.plot(self.x, self.curves[key], **self.style[key])

        if legend is True:
            # choose nb of columns in legend
            if len(self.curves.columns) > 4:
                legend_ncol = 2
            else:
                legend_ncol = 1
            plt.figlegend(fontsize='small', ncol=legend_ncol)

        # paint in black y=0 and x=0
        if black_axes is True:
            _black_axes(self.ax)
        # set y boudaries without taking outliers into account
        if outlier_thresh is not None:
            self.ignore_outliers(outlier_thresh)
        # give more space to y label
        box = self.ax.get_position()
        self.ax.set_position([box.x0+box.x0*0.1, box.y0,
                              box.width, box.height])

    def plot_median(self, style={}, envelope=False):
        """ """
        if 'linestyle' in style:
            style.pop('linestyle')
        quartiles = self.curves.quantile([0.25, 0.5, 0.75], axis=1).T
        plt.plot(self.x, quartiles[0.5],
                 linestyle='-', **style)
        if envelope is True:
            plt.plot(self.x, quartiles[[0.25, 0.75]],
                     linestyle='--', **style)

    def plot_mean(self, style={}, envelope=False):
        """ """
        if 'linestyle' in style:
            style.pop('linestyle')
        std = self.curves.std(axis=1).T
        mean = self.curves.mean(axis=1).T
        plt.plot(self.x, mean, linestyle='-', **style)
        if envelope is True:
            plt.plot(self.x, mean + std, linestyle='--', **style)
            plt.plot(self.x, mean - std, linestyle='--', **style)

    def ignore_outliers(self, outlier_thresh):
        """ Set y boudaries without taking outliers into account. """
        # seek outliers
        points = np.array([c['y'] for c in self.curves.values()])
        outliers = _is_outlier(points, outlier_thresh)
        # find min/max
        ymin = np.min(
            [np.array(c['y'])
             for c, keep in zip(self.curves.values(), ~outliers) if keep])
        ymax = np.max(
            [np.array(c['y'])
             for c, keep in zip(self.curves.values(), ~outliers) if keep])
        margin = (ymax - ymin)*.05
        ax.set_ylim(bottom=ymin-margin, top=ymax+margin)

    def write_csv(self, outputCSV, **kwargs):
        """ Save data.

        outputCSV : str or None
            File where to save the data. The file name must contain {score}
            instead of the score name.
        **kwargs :
            Keys of these arguments must be contained in the name of the
            output file and will be replaced by the corresponding values.

        """
        # check compulsory keys in file name
        for key in kwargs.keys():
            if "{" + key + "}" not in outputCSV:
                raise evt.EvaltoolsError(
                    "'{" + key + "}' must be " +
                    "contained in the name of the output csv file !!!")

        self.curves.to_csv(outputCSV.format(**kwargs), sep=' ', na_rep='nan',
                           float_format='%g', header=True, index=True)


def plot_meanTimeScores(
        objects, score_list, labels=None, colors=None,
        linestyles=None, markers=None, outputFile=None, suptitle="", title="",
        score_names=None, file_formats=['png'], min_nb_sta=10,
        availability_ratio=0.75, outputCSV=None,
        xlabel='Forecast time (hour)', black_axes=False, outlier_thresh=None,
        nb_of_minor_ticks=(1, 2)):
    """ Plot the mean of time scores for each forecast time.

    This function is based on Evaluator.meanTimeScores method.
    Scores are first computed for each time along available stations. Then,
    the mean of these scores is taken for each forecast time.

    On each plots, there will be as many lines as there are objects given in
    the object list.

    Parameters
    ----------
    objects : list of evaltools.Evaluator objects
        Evaluator objects used for plotting.
    score_list : list of str
        List of computed scores.
    labels : None or list of str
        List of labels for the legend.
    colors : None or list of str
        Line colors corresponding to each objects.
    linestyles : None or list of str
        Line styles corresponding to each objects.
    markers : None or list of str
        Line markers corresponding to each objects.
    outputFile : None or str
        File where to save the plots (without extension). The file name
        must contain {score} instead of the score name. If None, the figure
        is shown in a popping window.
    suptitle : str
        Suptitle for the plots.
    title : str
        Title for the plots. It must contain {score} instead of the
        score name.
    score_names : dict or None
        Dictionary with keys containing the elements of score_list and with
        values the strings to in the title instead of the default scores
        names. If None, score names are written as in score_list.
    file_formats : list of str
        List of file extensions.
    min_nb_sta : int
        Minimal number of station values available in both obs and
        sim required per datetime to compute the scores (before applying
        mean for each forecast time).
    availability_ratio : float
        Minimal rate of data (computed scores per time) available on the
        period required per forecast time to compute the mean scores.
    outputCSV : str or None
        File where to save the data. The file name must contain {score}
        instead of the score name.
    xlabel : str
        Label for the x axis.
    black_axes : bool
        If true, y=0 and x=0 lines are painted in black.
    outlier_thresh : scalar or None
        If not None, it correspond to the threshold used in
        evaltools.plotting._is_outlier to determine if a model is an outlier.
        If outliers are detected, y boudaries do not take them into account.
    nb_of_minor_ticks : tuple of int
        Number of minor ticks for x and y axes.

    """
    # check if objects are defined on the same period and same stations
    _checkPeriod(objects)
    _checkStationList(objects)

    # define default parameters
    labels, colors, linestyles, markers, legend_ncol, seriesType = \
        _defaultParams(objects, labels, colors, linestyles, markers)

    # define default score names
    if score_names is None:
        score_names = dict(zip(score_list, score_list))

    # abcissa
    X = range(24*objects[0].forecastHorizon)

    # scores computing
    stats = []
    for obj in objects:
        stats.append(obj.meanTimeScores(
            score_list,
            min_nb_sta=min_nb_sta,
            availability_ratio=availability_ratio))

    # plotting
    for sc in score_list:
        curves = _Curves(X)
        for Y, l, c, ls, m in zip(stats, labels, colors, linestyles, markers):
            curves.add_curve(l, Y[sc][:len(X)],
                             style={'label': l, 'color': c,
                                    'linestyle': ls, 'marker': m})
        curves.plot(black_axes=black_axes, outlier_thresh=outlier_thresh)
        plt.xlabel(xlabel)
        plt.ylabel(score_names[sc])
        plt.suptitle(suptitle, fontweight='bold', x=0.25, y=0.98)
        plt.title(title.format(score=score_names[sc]), loc='left', x=0.05)
        plt.tight_layout()
        plt.xlim(left=X[0], right=X[-1])
        major_ticks = np.arange(0, len(X), 12)
        minor_ticks = np.arange(0, len(X), 3)
        curves.ax.set_xticks(major_ticks)
        curves.ax.set_xticks(minor_ticks, minor=True)

        _set_minor_ticks(*nb_of_minor_ticks)

        _save_figure(outputFile, file_formats, score=sc)

        # save data
        if outputCSV is not None:
            curves.write_csv(outputCSV.format(score=sc))


def plot_medianStationScores(
        objects, score_list, labels=None, colors=None, linestyles=None,
        markers=None, outputFile=None, suptitle="", title="",
        score_names=None, file_formats=['png'], availability_ratio=0.75,
        min_nb_sta=10, outputCSV=None, xlabel='Forecast time (hour)',
        black_axes=False, outlier_thresh=None, nb_of_minor_ticks=(1, 2)):
    """ Plot the median of time scores for each forecast time.

    This function is based on Evaluator.medianStationScores method.
    Scores are first computed for each station at every times. Then,
    the median of these scores is taken for each forecast time.

    On each plots, there will be as many lines as there are objects given in
    the object list.

    Parameters
    ----------
    objects : list of evaltools.Evaluator objects
        Evaluator objects used for plotting.
    score_list : list of str
        List of computed scores.
    labels : None or list of str
        List of labels for the legend.
    colors : None or list of str
        Line colors corresponding to each objects.
    linestyles : None or list of str
        Line styles corresponding to each objects.
    markers : None or list of str
        Line markers corresponding to each objects.
    outputFile : None or str
        File where to save the plots (without extension). The file name
        must contain {score} instead of the score name. If None, the figure
        is shown in a popping window.
    suptitle : str
        Suptitle for the plots.
    title : str
        Title for the plots. It must contain {score} instead of the
        score name.
    score_names : dict
        Dictionary with keys containing the elements of score_list and as
        values the strings writen in the title instead of the default scores
        names.
    file_formats : list of str
        List of file extensions.
    availability_ratio : float
        Minimal rate of data available on the period required per forecast
        time to compute the scores for each station.
    min_nb_sta : int
        Minimal number of stations required to compute the median of
        the scores.
    outputCSV : str or None
        File where to save the data. The file name must contain {score}
        instead of the score name.
    xlabel : str
        Label for the x axis.
    black_axes : bool
        If true, y=0 and x=0 lines are painted in black.
    outlier_thresh : scalar or None
        If not None, it correspond to the threshold used in
        evaltools.plotting._is_outlier to determine if a model is an outlier.
        If outliers are detected, y boudaries do not take them into account.
    nb_of_minor_ticks : tuple of int
        Number of minor ticks for x and y axes.

    """
    # check if objects are defined on the same period and same stations
    _checkPeriod(objects)
    _checkStationList(objects)

    # define default parameters
    labels, colors, linestyles, markers, legend_ncol, seriesType = \
        _defaultParams(objects, labels, colors, linestyles, markers)

    # define default score names
    if score_names is None:
        score_names = dict(zip(score_list, score_list))

    # abcissa
    X = range(24*objects[0].forecastHorizon)

    # scores computing
    stats = []
    for obj in objects:
        stats.append(obj.medianStationScores(
            score_list,
            availability_ratio=availability_ratio,
            min_nb_sta=min_nb_sta))

    # plotting
    for sc in score_list:
        curves = _Curves(X)
        for Y, l, c, ls, m in zip(stats, labels, colors, linestyles, markers):
            curves.add_curve(l, Y[sc][:len(X)],
                             style={'label': l, 'color': c,
                                    'linestyle': ls, 'marker': m})
        curves.plot()
        plt.xlabel(xlabel)
        plt.ylabel(score_names[sc])
        plt.suptitle(suptitle, fontweight='bold', x=0.25, y=0.98)
        plt.title(title.format(score=score_names[sc]), loc='left', x=0.05)
        plt.tight_layout()
        plt.xlim(left=X[0], right=X[-1])
        major_ticks = np.arange(0, len(X), 12)
        minor_ticks = np.arange(0, len(X), 3)
        curves.ax.set_xticks(major_ticks)
        curves.ax.set_xticks(minor_ticks, minor=True)

        _set_minor_ticks(*nb_of_minor_ticks)

        _save_figure(outputFile, file_formats, score=sc)

        # save data
        if outputCSV is not None:
            curves.write_csv(outputCSV.format(score=sc))


def plot_timeScores(
        objects, score_list, term, labels=None, colors=None,
        linestyles=None, markers=None, outputFile=None, suptitle="", title="",
        score_names=None, file_formats=['png'], min_nb_sta=10, outputCSV=None,
        xticking='auto', black_axes=False, outlier_thresh=None, 
        nb_of_minor_ticks=(2, 2)):
    """ Plot time scores at a chosen time for each forecast day.

    This function is based on Evaluator.timeScores method.
    On each plots, there will be as many lines as there are objects the
    object list.

    Parameters
    ----------
    objects : list of evaltools.Evaluator objects
        Evaluator objects used for plotting.
    score_list : list of str
        List of computed scores.
    term : int.
        Integer corresponding to the chosen term used for plotting.
        If the series type of the objects is 'hourly', it refers
        to the forecast time (for example between 0 and 95 if the forecast
        horizon is 4 days). If the series type of the objects is 'daily',
        it refers to the forcast day (for example between 0 and 3 if the
        forecast horizon is 4 days).
    labels : None or list of str
        List of labels for the legend.
    colors : None or list of str
        Line colors corresponding to each objects.
    linestyles : None or list of str
        Line styles corresponding to each objects.
    markers : None or list of str
        Line markers corresponding to each objects.
    outputFile : None or str
        File where to save the plots (without extension). The file name
        must contain {score} instead of the score name. If None, the figure
        is shown in a popping window.
    suptitle : str
        Suptitle for the plots.
    title : str
        Title for the plots. It must contain {score} instead of the
        score name.
    score_names : dict
        Dictionary with keys containing the elements of score_list and as
        values the strings writen in the title instead of the default scores
        names.
    file_formats : list of str
        List of file extensions.
    min_nb_sta : int
        Minimal number of station values available in both obs and sim
        required per datetime to compute the scores.
    outputCSV : str or None
        File where to save the data. The file name must contain {score}
        instead of the score name.
    xticking : str
        Defines the method used to set x ticks. It can be 'auto' (automatic),
        'mondays' (a tick every monday) or 'daily' (a tick everyday).
    black_axes : bool
        If true, y=0 and x=0 lines are painted in black.
    outlier_thresh : scalar or None
        If not None, it correspond to the threshold used in
        evaltools.plotting._is_outlier to determine if a model is an outlier.
        If outliers are detected, y boudaries do not take them into account.
    nb_of_minor_ticks : tuple of int
        Number of minor ticks for x and y axes.

    """
    # check if objects are defined on the same period and same stations
    _checkPeriod(objects)
    _checkStationList(objects)

    # define default parameters
    labels, colors, linestyles, markers, legend_ncol, seriesType = \
        _defaultParams(objects, labels, colors, linestyles, markers)

    # define whether if term is the forecast day or forecast time
    if seriesType == 'hourly':
        # forecast day corresponding to term
        day = term//24
        # hour of the day corresponding term
        t = term % 24
    elif seriesType == 'daily':
        day = term

    # define default score names
    if score_names is None:
        score_names = dict(zip(score_list, score_list))

    # scores computing
    stats = []
    for obj in objects:
        Y = obj.timeScores(score_list,
                           min_nb_sta=min_nb_sta)['D{day}'.format(day=day)]
        if seriesType is 'hourly':
            idx = Y.index.time == time(t, 0)
            Y = Y[idx]
        stats.append(Y)

    # plotting
    for sc in score_list:
        if seriesType is 'hourly':
            Xticks = np.unique(stats[0].index.date)
            X = [datetime.combine(x, time(t)) for x in Xticks]
        elif seriesType is 'daily':
            Xticks = stats[0].index
            X = Xticks
        curves = _Curves(X)

        for Y, l, c, ls, m, obj in zip(stats, labels, colors, linestyles,
                                       markers, objects):
            if Y[sc].isna().all():
                print("*** all nan values for {sc} of {label}".format(
                    sc=sc, label=l))
            else:
                curves.add_curve(l, Y[sc][:len(X)],
                                 style={'label': l, 'color': c,
                                        'linestyle': ls, 'marker': m})
        curves.plot()

        plt.ylabel(score_names[sc])
        plt.suptitle(suptitle, fontweight='bold', x=0.25, y=0.98)
        plt.title(title.format(score=score_names[sc]), loc='left', x=0.05)
        plt.tight_layout()

        # set xticks
        if seriesType is 'hourly':
            Xticks = np.append(Xticks, Xticks[-1]+timedelta(days=1))
        curves.fig.autofmt_xdate(bottom=0.17, rotation=70, ha='right')
        if xticking == 'mondays':
            xt = Xticks[np.array([d.weekday() == 0 for d in Xticks])]
            curves.ax.set_xticks(xt)
            curves.ax.set_xticklabels([d.strftime('%Y-%m-%d') for d in xt])
            # curves.ax.set_xticklabels([d.strftime('%d/%m/%y') for d in xt])
        elif xticking == 'daily':
            curves.ax.set_xticks(np.array(Xticks))
            curves.ax.set_xticklabels([d.strftime('%Y-%m-%d')
                                       for d in Xticks])
            # curves.ax.set_xticklabels([d.strftime('%d/%m/%y')
            #                            for d in Xticks])
        else:
            locator = dates.AutoDateLocator(minticks=2)
            curves.ax.xaxis.set_major_locator(locator)
            hfmt = dates.AutoDateFormatter(locator)
            curves.ax.xaxis.set_major_formatter(hfmt)

        plt.xlim(left=Xticks[0], right=Xticks[-1])
        plt.tick_params(axis='x', which='major', labelsize='small')

        _set_minor_ticks(*nb_of_minor_ticks)

        _save_figure(outputFile, file_formats, score=sc)

        # save data
        if outputCSV is not None:
            curves.write_csv(outputCSV.format(score=sc))


def plot_quarterlyMedianScore(
        files, labels, colors, first_quarter=None, last_quarter=None,
        score='RMSE', outputFile=None, linestyles=None, markers=None,
        suptitle="", title="", thres=None, file_formats=['png'], ylabel=None,
        origin_zero=False, black_axes=False):
    """ Plot quarterly values saved in given files.

    This function is based on Evaluator.quarterlyMedianScore method.
    On each plots, there will be as many lines as there are file paths
    in <files> argument.

    Parameters
    ----------
    files : list of str
        Paths of files used for plotting (one file corresponds to one line
        on the plot).
    labels : list of str
        List of labels for the legend (length of the label list must be
        equal to the length of the file list).
    colors : list of str
        Line colors corresponding to each files.
    first_quarter : str
        String corresponding to the oldest plotted quarter.
    last_quarter : str
        String corresponding to the latest plotted quarter.
    score : str
        The score that should correspond to data stored in the files.
    outputFile : str
        File where to save the plot (without extension).
    linestyles : None or list of str
        Line styles corresponding to each objects.
    markers : None or list of str
        Line markers corresponding to each objects.
    suptitle : str
        Suptitle for the plots.
    title : str
        Title for the plots.
    thres : None or float
        If not None, a horizontal yellow line is plotted with
        equation y = thres.
    file_formats : list of str
        List of file extensions.
    ylabel : str
        Ordinate axis label.
    origin_zero : bool
        If True, minimal value of y axis is set to zero.
    black_axes : bool
        If true, y=0 and x=0 lines are painted in black.

    """
    # define linestyles in case of None
    if linestyles is None:
        linestyles = ['-']*len(files)

    # define markers in case of None
    if markers is None:
        markers = [default_marker]*len(files)

    # find first and last quarters in case of None
    if last_quarter is None or first_quarter is None:
        dfs = [pd.read_csv(f, sep=' ', index_col=0) for f in files]
        quarter_list = []
        for df in dfs:
            quarter_list.extend([evt.quarter.Quarter.from_string(q)
                                 for q in df.index])
        quarter_list.sort()
        if first_quarter is None:
            first_quarter = quarter_list[0].string
        if last_quarter is None:
            last_quarter = quarter_list[-1].string

    end_quarter = evt.quarter.Quarter.from_string(last_quarter)
    start_quarter = evt.quarter.Quarter.from_string(first_quarter)
    all_quarters = end_quarter.range(start_quarter)

    Ys = {}
    for kpiFile, lab in zip(files, labels):
        df = pd.read_csv(kpiFile, sep=' ', index_col=0)
        Ys[lab] = pd.DataFrame(index=[q.__repr__() for q in all_quarters],
                               columns=[score])
        Ys[lab].update(df)

    X = range(len(all_quarters))

    plt.close('all')
    fig = plt.figure(figsize=figsize)
    if thres is not None:
        plt.plot(X, [thres]*len(X), color='#FFA600')
    for lab, c, ls, m in zip(labels, colors, linestyles, markers):
        Y = Ys[lab].astype('float')
        idx = (~np.isnan(Y)).values[:, 0]
        plt.plot(X, Y, color=c, linestyle=ls, marker=m, label=lab)
    if ylabel is not None:
        plt.ylabel(ylabel)
    else:
        plt.ylabel(Ys[labels[0]].keys()[0])
    plt.suptitle(suptitle, fontweight='bold', x=0.25, y=0.98)
    plt.title(title, loc='left', x=0.05)
    plt.tight_layout()
    plt.grid(True, ls='-', c='#B0B0B0')
    plt.figlegend()
    plt.xlim(left=X[0], right=X[-1])
    if origin_zero is True:
        plt.ylim(bottom=0)
    plt.xticks(X, Ys[lab].index)
    plt.locator_params(axis='y', nbins=10, min_n_ticks=10)
    fig.autofmt_xdate(bottom=0.2, rotation=80, ha='right')

    # paint in black y=0 and x=0
    if black_axes is True:
        _black_axes(fig.get_axes()[0])

    # save figure
    _save_figure(outputFile, file_formats)


def plot_stationScores(
        obj, score, forecastDay=0, outputFile=None, title="",
        bbox=[-26, 46, 28, 72], file_formats=['png'], point_size=5,
        higher_above=True, order_by=None, availability_ratio=0.75, vmin=None,
        vmax=None, cmap=cm.jet, rivers=False, outputCSV=None, interp2D=False,
        sea_mask=False, land_mask=False, boundary_resolution='10m'):
    """ Plot scores per station on a map.

    This function is based on Evaluator.stationScores.

    Each station is drawn with a circle colored according to
    it score value.

    Parameters
    ----------
    obj : evaltools.Evaluator object
        Object used for plotting.
    score : str
        Computed score.
    forecastDay : int.
        Integer corresponding to the chosen forecastDay used for plotting.
    outputFile : str or None
        File where to save the plots (without extension). If None, the figure
        is shown in a popping window.
    title : str
        Title for the plots. It must contain {score} instead of the
        score name.
    bbox : list of floats
        Bounding box of plotting area [min_lon, max_lon, min_lat, max_lat].
    file_formats : list of str
        List of file extensions.
    point_size : float
        Point size (as define in matplotlib.pyplot.scatter).
    higher_above : bool
        If True, stations with higher score are plotted above. If False,
        stations with lower score are plotted above.
    order_by : str
        Vertically order point according to this argument. It must be a
        column of the Evaluator object stations attribute.
    availability_ratio : float
        Minimal rate of data available on the period required per forecast
        term to compute the mean scores.
    vmin, vmax : None or scalar
        Min and max values for the legend colorbar. If None, the respective
        min and max of the score values are used.
    cmap : matplotlib.colors.Colormap object
        Colors used for plotting (default: matplotlib.cm.jet).
    rivers : bool
        If True, rivers and lakes are drawn.
    outputCSV : str or None
        File where to save the data. The file name can contain {score}
        instead of the score name.
    interp2D : bool
        If True, a 2D linear interplation is performed on scores values.
    sea_mask : bool
        If True, scores ought to be drawn over sea are masked.
    land_mask : bool
        If True, scores ought to be drawn over land are masked.
    boundary_resolution : str
        Resolution of coastlines and boundary lines. It can be '10m',
        '50m' or '110m'.

    """
    try:
        import cartopy.crs as ccrs
        import cartopy.feature as cfeature
    except ImportError as e:
        print("Module cartopy is required to use plot_timeScores function.")
    else:
        # compute scores
        df = obj.stationScores(
            [score],
            availability_ratio=availability_ratio)['D{}'.format(forecastDay)]

        # sort values by plotted values (or metadata if order_by is not None)
        df.sort_values(by=score, ascending=higher_above, inplace=True)
        if order_by is not None:
            if np.in1d(order_by, obj.stations.columns):
                df = df.merge(obj.stations[[order_by]], how='left',
                              left_index=True, right_index=True, sort=False)
                df.sort_values(ascending=higher_above, by=order_by,
                               inplace=True)
            else:
                print(("*** Values can't be ordered by {} since " +
                       "this variable is not in the object's " +
                       "metadata").format(order_by))
        df = df[score]

        # plotting
        plt.close('all')
        fig = plt.figure(figsize=figsize)

        ax = plt.axes(projection=ccrs.PlateCarree())
        # ax = plt.axes(projection=ccrs.Mercator())
        ax.coastlines(resolution=boundary_resolution)
        # ax.add_feature(cfeature.BORDERS, edgecolor='lightgrey')
        ax.add_feature(cfeature.NaturalEarthFeature(
            'cultural', 'admin_0_boundary_lines_land',
            scale=boundary_resolution, edgecolor='lightgrey',
            facecolor='none'))

        # land/sea mask
        if land_mask is True:
            ax.add_feature(cfeature.NaturalEarthFeature('physical', 'land',
                                                        boundary_resolution,
                                                        edgecolor='k',
                                                        facecolor='w'))
        if sea_mask is True:
            ax.add_feature(cfeature.NaturalEarthFeature('physical', 'ocean',
                                                        boundary_resolution,
                                                        edgecolor='k',
                                                        facecolor='w'))

        if rivers is True:
            # ax.add_feature(cfeature.RIVERS)
            ax.add_feature(cfeature.NaturalEarthFeature(
                'physical', 'lakes', '50m',
                edgecolor='face', facecolor=np.array((152, 183, 226)) / 256.))
            # ax.add_feature(cfeature.LAKES, alpha=0.5)
            ax.add_feature(
                cfeature.NaturalEarthFeature(
                    'physical', 'rivers_lake_centerlines', '50m',
                    edgecolor=np.array((152, 183, 226)) / 256.,
                    facecolor='none'),
                alpha=0.5)

        # set boundary box
        x1, x2, y1, y2 = bbox
        proj_bbox = ccrs.Mercator().transform_points(
            ccrs.PlateCarree(), np.array(bbox[0:2]), np.array(bbox[2:4]))
        ax.set_extent(bbox, crs=ccrs.PlateCarree())
        # ax.set_xlim(left=proj_bbox[0, 0], right=proj_bbox[1, 0])
        # ax.set_ylim(bottom=proj_bbox[0, 1], top=proj_bbox[1, 1])

        if interp2D is True:
            from scipy.interpolate import griddata
            grid_lat, grid_lon = np.mgrid[y1:y2:1000j, x1:x2:1000j]
            interp_data = griddata(
                np.array(obj.stations[['lat', 'lon']].loc[df.index]),
                df.values, (grid_lat, grid_lon),
                method='linear')
            im = ax.imshow(interp_data, extent=bbox,
                           transform=ccrs.PlateCarree(),
                           cmap=cmap, vmin=vmin, vmax=vmax)
            # im = ax.pcolormesh(grid_lon[0, :], grid_lat[:, 0], interp_data,
            #                    transform=ccrs.Mercator())
            plt.colorbar(im)
        else:
            # scatter plot
            x = np.array(obj.stations.lon.loc[df.index])
            y = np.array(obj.stations.lat.loc[df.index])
            scatter = ax.scatter(
                x, y, marker='o', s=point_size, c=df, cmap=cmap,
                edgecolors=None, vmin=vmin, vmax=vmax, zorder=10,
                transform=ccrs.PlateCarree())
            plt.colorbar(scatter)
        plt.title(title)

        # annotate number of stations
        plt.annotate('({av} processed stations over {total})'.format(
                         av=sum(df.notna()), total=df.shape[0]),
                     xy=(0, 0),
                     xytext=(0, -10),
                     xycoords='axes fraction',
                     textcoords='offset points',
                     va='top',
                     fontsize='small', fontstyle='italic', color='#7A7A7A')

        # save figure
        _save_figure(outputFile, file_formats)

        # save data
        if outputCSV is not None:
            csv_data = obj.stations[['lat', 'lon']].loc[df.index]
            csv_data[score] = df
            csv_data.index.name = 'station'
            csv_data.sort_values(by=score)
            csv_data.to_csv(outputCSV.format(score=score), sep=' ',
                            na_rep='nan', float_format='%g', header=True,
                            index=True)


def plot_taylorDiagram(
        objects, forecastDay=0, colors=None, markers=None,
        point_size=100, outputFile=None, labels=None, suptitle="", title="",
        file_formats=['png'], threshold=0.75, outputCSV=None):
    """ Taylor diagram.

    This function is based on Evaluator.FDscores method.
    Pearson correlation and variance ratio are first computed from all data
    of a choosen forecast day (values for all station at all times are
    considered as a simple 1D array).

    Parameters
    ----------
    objects : list of evaltools.Evaluator objects
        Evaluator objects used for plotting.
    forecastDay : int
        Forecast day for which to use data to compute the Taylor diagram.
    colors : None or list of str
        Marker colors corresponding to each objects.
    markers : None or list of str
        Marker shapes corresponding to each objects.
    point_size : float
        Point size (as define in matplotlib.pyplot.scatter).
    outputFile : str
        File where to save the plot (without extension).
    labels : list of str
        List of labels for the legend.
    suptitle : str
        Diagram suptitle.
    title : str
        Diagram title.
    file_formats : list of str
        List of file extensions.
    threshold : int or float
        Minimal number (if type(threshold) is int) or minimal rate
        (if type(threshold) is float) of data available in both obs and
        sim required to compute the scores.
    outputCSV : str or None
        File where to save the data.

    """
    # check if objects are defined on the same period and same stations
    _checkPeriod(objects)
    _checkStationList(objects)

    # define default parameters
    linestyles = None
    labels, colors, linestyles, markers, legend_ncol, seriesType = \
        _defaultParams(objects, labels, colors, linestyles, markers,
                       default_marker='^')

    # plotting
    plt.close('all')
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    ax.set_xlim(0, 1.6)
    ax.set_ylim(0, 1.6)
    major_ticks = np.arange(0, 1.6, 1)
    minor_ticks = np.arange(0, 1.6, 0.1)
    ax.set_xticks(major_ticks)
    ax.set_xticks(minor_ticks, minor=True)
    ax.set_yticks(major_ticks)
    ax.set_yticks(minor_ticks, minor=True)
    ax.set_aspect('equal')

    refstd = 1
    smax = 1.6
    levels = 10
    xs, ys = np.meshgrid(np.linspace(0, smax), np.linspace(0, smax))
    rms = np.sqrt((refstd-xs)**2 + ys**2)
    contours = ax.contour(xs, ys, rms, levels, colors='k',
                          linestyles='dotted', linewidths=0.5)

    corr_steps = np.concatenate((np.arange(10)/10., [0.95, 0.99]))
    m = 0.02
    for corr in corr_steps:
        ang = np.arccos(corr)
        x = 1.5*np.cos(ang)
        y = 1.5*np.sin(ang)
        plt.plot([0, x], [0, y], 'c', lw=0.5)
        if corr != 0.:
            plt.text(x+m*np.cos(ang), y+m*np.sin(ang), corr, color='b',
                     rotation=(ang-np.pi/2)*180/np.pi,
                     verticalalignment='center', horizontalalignment='center')
    plt.text(1.6*np.sqrt(2)/2, 1.6*np.sqrt(2)/2, "correlation", color='b',
             rotation=-45, verticalalignment='center',
             horizontalalignment='center', size='x-large')

    for r in [1., 1.5]:
        theta = np.linspace(0, np.pi/2)
        x = np.cos(theta)*r
        y = np.sin(theta)*r
        ax.plot(x, y, c='#FF3FFF', linestyle='-')

    plt.xlabel(r"Standard deviation ratio ($\sigma_{mod} / \sigma_{obs}$)")
    plt.ylabel(r"Standard deviation ratio ($\sigma_{mod} / \sigma_{obs}$)")

    # plot model points
    csv_data = pd.DataFrame(columns=['corr', 'SDratio', 'CRMSE'])
    for obj, color, marker, lab in zip(objects, colors, markers, labels):
        df = obj.FDscores(score_list=['Ratio', 'PearsonR', 'CRMSE'],
                          forecast_days=[forecastDay], threshold=threshold)
        corr = df['PearsonR'].loc['D{}'.format(forecastDay)]
        sd_ratio = df['Ratio'].loc['D{}'.format(forecastDay)]
        x = corr*sd_ratio
        y = np.sin(np.arccos(corr))*sd_ratio
        ax.scatter(x, y, c=color, marker=marker, label=lab, s=point_size,
                   zorder=10)
        csv_data = pd.concat([csv_data,
                              pd.DataFrame({'corr': corr,
                                            'SDratio': sd_ratio,
                                            'CRMSE': df['CRMSE'].loc[
                                                'D{}'.format(forecastDay)]},
                                           index=[lab])],
                             sort=False)

    plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    # plt.figlegend(fontsize='small')
    plt.suptitle(suptitle, horizontalalignment='center', fontweight='bold')
    plt.title(title)
    # plt.tight_layout()

    # Shrink current axis by 20%
    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

    # save figure
    _save_figure(outputFile, file_formats)

    # save data
    if outputCSV is not None:
        csv_data.index.name = 'model'
        csv_data.to_csv(outputCSV, sep=' ', na_rep='nan', float_format='%g',
                        header=True, index=True)


def plot_scoreQuartiles(
        objects, xscore, yscore, colors=None, forecastDay=0, title="",
        outputFile=None, labels=None, availability_ratio=0.75, min_nb_sta=10,
        file_formats=['png'], outputCSV=None, invert_xaxis=False,
        invert_yaxis=False, xmin=None, xmax=None, ymin=None, ymax=None,
        black_axes=False):
    """ Scatter plot of median station scores.

    Plot the median of the station scores surrounded by a rectangle
    correponding to the first and third quartiles. This chart is
    based on Evaluator.stationScores method.

    Parameters
    ----------
    objects : list of evaltools.Evaluator objects
        Evaluator objects used for plotting.
    xscore : str
        Score used for the x axis.
    yscore : str
        Score used for the y axis.
    colors : None or list of str
        Marker colors corresponding to each objects.
    forecastDay : int
        Forecast day used for score computing.
    title : str
        Chart title.
    outputFile : str
        File where to save the plot (without extension).
    labels : list of str
        List of objects labels for the legend.
    availability_ratio : float
        Minimal rate of data available on the period required per forecast
        day to compute the scores for each station.
    min_nb_sta : int
        Minimal number of stations required to compute the quartiles of
        the scores.
    file_formats : list of str
        List of file extensions.
    outputCSV : str or None
        File where to save the data.
    invert_xaxis, invert_yaxis : bool
        If True, the axis is inverted.
    xmin, xmax, ymin, ymax : None or scalar
        Limits of the axes.
    black_axes : bool
        If true, y=0 and x=0 lines are painted in black.

    """
    # check if objects are defined on the same period and same stations
    _checkPeriod(objects)
    _checkStationList(objects)

    # build 'labels' from objects 'model' attribute if 'labels' is None
    if labels is None:
        labels = [obj.model for obj in objects]

    # define objects colors in case of None
    if colors is None:
        colors = []
        for obj in objects:
            colors.append(obj.color)

    # score computing
    res = {}
    for obj, lab in zip(objects, labels):
        stats = obj.stationScores(
            score_list=[xscore, yscore],
            availability_ratio=availability_ratio)['D{}'.format(forecastDay)]
        if (np.sum(~stats[xscore].isna(), axis=0) > min_nb_sta):
            quartiles = np.nanpercentile(stats, q=[25, 50, 75], axis=0)
            quartiles = pd.DataFrame(quartiles, index=[25, 50, 75],
                                     columns=stats.columns)
            res[lab] = quartiles
        else:
            print("Warning: Availability criteria not fulfilled " +
                  "for {} !!!".format(lab))

    # plotting
    plt.close('all')
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(1, 1, 1)
    plt.title(title)
    for obj, lab, col in zip(objects, labels, colors):
        rect = plt.Rectangle(
            xy=tuple(res[lab][[xscore, yscore]].loc[25]),
            width=res[lab][xscore].loc[75]-res[lab][xscore].loc[25],
            height=res[lab][yscore].loc[75]-res[lab][yscore].loc[25],
            edgecolor=col,
            fc='none',
            lw=2)
        ax.add_patch(rect)
        plt.gca().add_patch(rect)
        ax.scatter([res[lab][xscore].loc[50]], [res[lab][yscore].loc[50]],
                   s=50, c=col, marker='o', label=lab)

    ax.set_xlim(left=xmin, right=xmax)
    ax.set_ylim(bottom=ymin, top=ymax)
    ax.grid(True)
    ax.set_axisbelow(True)
    leg = ax.legend(loc='best', scatterpoints=1, prop={'size': 10})
    leg.draw_frame(False)
    ax.set_xlabel(xscore, fontsize=12)
    ax.set_ylabel(yscore, fontsize=12)
    if invert_xaxis is True:
        plt.gca().invert_xaxis()
    if invert_yaxis is True:
        plt.gca().invert_yaxis()

    # paint in black y=0 and x=0
    if black_axes is True:
        _black_axes(ax)

    # save figure
    _save_figure(outputFile, file_formats)

    # save data
    if outputCSV is not None:
        with open(outputCSV, 'w') as f:
            for lab in labels:
                f.write(lab+'\n')
                f.write(res[lab].to_string())
                f.write('\n\n')


def plot_comparisonScatterPlot(
        score, xobject, yobject, forecastDay=0, title="", xlabel=None,
        ylabel=None, availability_ratio=0.75, nb_outliers=5, outputFile=None,
        file_formats=['png'], outputCSV=None, black_axes=False,
        color_by=None):
    """ Scatter plot to compare two Evaluator objects.

    One score is calculated for two Evaluator objects and used as xy
    coordinates. Points are colored according to the density of points.

    Parameters
    ----------
    score : str
        Computed score.
    xobject, yobject : evaltools.Evaluator object
        Objects used for plotting.
    forecastDay : int
        Integer corresponding to the chosen forecastDay used for plotting.
    title : str
        Title for the plots. It must contain {score} instead of the
        score name.
    xlabel, ylabel : str
        Labels corresponding to xobject and yobject used for axis labels.
    availability_ratio : float
        Minimal rate of data available on the period required per forecast
        day to compute station scores.
    outputFile : str or None
        File where to save the plot (without extension). The file name
        must contain {score} instead of the score name. If None, the figure
        is shown in a popping window.
    file_formats : list of str
        List of file extensions.
    outputCSV : str or None
        File where to save the data.
    black_axes : bool
        If true, y=0 and x=0 lines are painted in black.
    color_by : None or dictionary
        Dictionary with keys corresponding to station names and
        values corresponding to colors.

    """
    # check if objects are defined on the same period and same stations
    _checkPeriod([xobject, yobject])
    _checkStationList([xobject, yobject])

    # build 'labels' from objects 'model' attribute if 'labels' is None
    if xlabel is None:
        xlabel = xobject.model
    if ylabel is None:
        ylabel = yobject.model

    # score computing
    stations = np.intersect1d(xobject.stations.index, yobject.stations.index)
    stats = pd.DataFrame(index=stations, columns=[xlabel, ylabel])
    for obj, lab in zip([xobject, yobject], [xlabel, ylabel]):
        df = obj.stationScores(
            score_list=[score],
            availability_ratio=availability_ratio)[
                'D{}'.format(forecastDay)][score]
        df.name = lab
        df = pd.DataFrame(df)
        stats.update(df)

    # min max values
    minval = np.nanmin(stats.values)
    maxval = np.nanmax(stats.values)

    # distance to y = x
    stats['d'] = abs(stats[xlabel]-stats[ylabel])/np.sqrt(2)

    # define point color in stats['z']
    if color_by is None:
        # remove nan to compute gaussian_kde
        idx_not_nan = ~stats.isna().any(axis=1)
        xy = stats.loc[idx_not_nan].values.T.astype('float')
        try:
            z = gaussian_kde(xy)(xy)
            z = pd.DataFrame(z, index=idx_not_nan[idx_not_nan].index,
                             columns=['z'])
            stats['z'] = np.nan
            stats.update(z)
            # sort points by density (=> densest points are plotted last)
            stats.sort_values(by='z', inplace=True)
        except np.linalg.linalg.LinAlgError:
            stats['z'] = 'blue'
    else:
        stats['z'] = pd.Series(color_by)
        stats.sort_values(by='z', inplace=True)

    # plotting
    plt.close('all')
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(1, 1, 1)
    ax.scatter(stats[xlabel], stats[ylabel], c=stats['z'], s=50,
               edgecolor='', zorder=10)

    # outliers highlighting
    stats.sort_values(by='d', ascending=False, inplace=True)
    outliers = stats[:nb_outliers]
    ax.scatter(outliers[xlabel], outliers[ylabel], s=100, facecolors='none',
               edgecolors='0.5')

    # line y = x
    XF = np.linspace(minval, maxval, 1000)
    plt.plot(XF, XF, 'k--', zorder=11)

    # layout
    plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.axis('equal')
    plt.grid(True)
    ax.set_axisbelow(True)
    if minval < 0 < maxval:
        plt.axhline(y=0, color='k')
        plt.axvline(x=0, color='k')

    # paint in black y=0 and x=0
    if black_axes is True:
        fig.canvas.draw()
        _black_axes(ax)

    # save figure
    _save_figure(outputFile, file_formats)

    # save data
    if outputCSV is not None:
        with open(outputCSV, 'w') as f:
            stats.__delitem__('z')
            stats = stats.merge(xobject.stations, how='left', left_index=True,
                                right_index=True, sort=False)
            f.write(stats.to_string())


def plot_significantDifferences(
        score_list, former_objects, later_objects, forecastDay=0, title="",
        xlabels=None, ylabels=None, availability_ratio=0.75, outputFile=None,
        file_formats=['png']):
    """ Chart showing the significativity of differences between simulations.

    Parameters
    ----------
    score_list : list of str
        List of scores for which investigating differences.
    former_objects, later_objects : list of evaltools.Evaluator objects
        Two list of objects that are compared.
    forecastDay : int
        Integer corresponding to the chosen forecastDay used for plotting.
    title : str
        Title for the plots. It must contain {score} instead of the
        score name.
    xlabels, ylabels : None list of str
        Labels for the x axis (equal to score_list if None) and labels
        for the y axis (corresponding to each object couple comparison
        and equal to the model name of the first object of each couple
        if None).
    availability_ratio : float
        Minimal rate of data available on the period required per forecast
        day to compute station scores.
    outputFile : str or None
        File where to save the plot (without extension). The file name
        must contain {score} instead of the score name. If None, the figure
        is shown in a popping window.
    file_formats : list of str
        List of file extensions.

    """
    # build 'labels' from objects 'model' attribute if 'labels' is None
    if xlabels is None:
        xlabels = score_list
    if ylabels is None:
        ylabels = [obj.model for obj in former_objects]

    percentlist = range(10, 100, 10)

    # tests performing
    difftests = {}
    for xobject, yobject, lab in zip(former_objects, later_objects, ylabels):
        difftests[lab] = {}
        # score computing
        stations = np.intersect1d(xobject.stations.index,
                                  yobject.stations.index)
        stats_x = xobject.stationScores(
            score_list=score_list,
            availability_ratio=availability_ratio)[
                'D{}'.format(forecastDay)].loc[stations]
        stats_y = yobject.stationScores(
            score_list=score_list,
            availability_ratio=availability_ratio)[
                'D{}'.format(forecastDay)].loc[stations]
        for sc in score_list:
            idx_not_nan = ~np.logical_or(stats_x[sc].isna(),
                                         stats_y[sc].isna())
            xval = stats_x[sc].loc[idx_not_nan]
            yval = stats_y[sc].loc[idx_not_nan]
            px = [scoreatpercentile(xval.loc[idx_not_nan], q)
                  for q in percentlist]
            py = [scoreatpercentile(yval, q)
                  for q in percentlist]
            chg = 0
            if sc in ['Bias', 'FracBias', 'MMB', 'FGE', 'RMSE', 'NMSE',
                      'CRMSE']:
                for k in range(len(px)):
                    if abs(py[k]) < abs(px[k]):
                        chg += 1
                    elif abs(py[k]) > abs(px[k]):
                        chg += -1
            else:
                for k in range(len(px)):
                    if py[k] > px[k]:
                        chg += 1
                    elif py[k] < px[k]:
                        chg += -1

            # Welch T-test on means (independant samples with != variance)
            # Mann-Whitney U-test (no gaussian hypotethis)
            v1, p1 = ttest_ind(xval, yval, equal_var=False)
            v2, p2 = mannwhitneyu(xval, yval)
            tpl = (len(xval), chg, p1, p2)
            difftests[lab][sc] = tpl

    # choosing colors according to the tests results
    xlab = []
    lx = len(score_list)
    ly = len(ylabels)
    matx = np.zeros((ly, lx))
    for s in range(ly):
        for i in range(lx):
            slab = ylabels[s]
            ilab = score_list[i]
            if (difftests[slab][ilab][2] <= 0.05 and
                    difftests[slab][ilab][3] <= 0.05):
                matx[s, i] = difftests[slab][ilab][1]
        xlab.append(slab + ' (#' + str(difftests[slab][ilab][0]) + ')')

    # plotting
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(1, 1, 1)
    plt.pcolor(matx, cmap='RdYlGn', vmin=-(len(percentlist) + 1),
               vmax=len(percentlist) + 1)
    colorbar = plt.colorbar()
    tick_x = np.arange(0.5, lx + 0.5)
    tick_y = np.arange(0.5, ly + 0.5)
    plt.xticks(tick_x, score_list, rotation=45, ha='right')
    plt.yticks(tick_y, xlab)
    # ax.set_aspect('equal', 'datalim')
    plt.title(title)
    plt.gcf().subplots_adjust(bottom=0.3, left=0.3)
    plt.tight_layout()

    # save figure
    _save_figure(outputFile, file_formats)


def plot_timeSeries(
        objects, station_list=None, start_end=None, forecastDay=0,
        colors=None, linestyles=None, markers=None, outputFile=None,
        labels=None, title="", ylabel='concentration', file_formats=['png'],
        outputCSV=None, ymin=None, ymax=None, xticking='auto', plot_type=None,
        envelope=False, black_axes=False, 
        nb_of_minor_ticks=(2, 2), obs_style={'color': 'k', 'alpha': 0.5}):
    """ Plot time series of obs and sim at a chosen forecast day.

    By default, there is one line per station and per object.
    Observations are taken from the fisrt object of the list (if of type
    Evaluator).

    Parameters
    ----------
    objects : list of evaltools.Evaluator or evaltools.Simulations objects
        Evaluator or Simulations objects used for plotting.
    station_list : None or list of str
        List of stations to display. If None, all stations of the
        stations of the first element of <objects> argument are
        processed.
    start_end : None or list of two datetime.date objects
        Boundary dates for abscissa.
    forecastDay : int
        Forecast day for which to plot the data.
    colors : None or list of str
        Line colors corresponding to each objects.
    linestyles : None or list of str
        Line styles corresponding to each objects.
    markers : None or list of str
        Line markers corresponding to each objects.
    outputFile : str or None
        File where to save the plots (without extension). If None, the figure
        is shown in a popping window.
    labels : None or list of str
        List of labels for the legend.
    title : str
        Title for the plot.
    ylabel : str
        Label for the y axis.
    file_formats : list of str
        List of file extensions.
    outputCSV : str or None
        File where to save the data. The file name must contain {model}
        instead of the model name.
    ymin, ymax : None or scalar
        Limits of the axes.
    xticking : str
        Defines the method used to set x ticks. It can be 'auto' (automatic),
        'mondays' (a tick every monday) or 'daily' (a tick everyday).
    quartiles : bool
        If True, draw quartiles instead of all stations separately.
    plot_type : str
        If 'median' the median of all stations values for a given time
        is plotted. If 'mean' the mean of all stations values for a given
        time is plotted. For any other value, all station values are plotted
        separately.
    envelope : bool
        Only use if plot_type is 'median' or 'mean'. If True, draw quartiles
        one and three around the median curve (case plot_type == 'median')
        or draw +/- the standard deviation around the mean curve (case
        plot_type == 'mean').
    black_axes : bool
        If true, y=0 and x=0 lines are painted in black.
    nb_of_minor_ticks : tuple of int
        Number of minor ticks for x and y axes.
    obs_style : dict
        Dictionary of arguments passed to pyplot.plot for observation
        curve. Default value is {'color': 'k', 'alpha': 0.5}.

    """
    # define default parameters
    labels, colors, linestyles, markers, legend_ncol, seriesType = \
        _defaultParams(objects, labels, colors, linestyles, markers)
    if station_list is None:
        station_list = reduce(np.intersect1d, [obj.simDF[0].columns
                                               for obj in objects])
    if not 'label' in obs_style:
        obs_style['label'] = 'observations'

    # plotting
    plt.close('all')
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(1, 1, 1)

    # first object in the list defines abscissa boundaries
    if start_end is None:
        start_end = [objects[0].startDate, objects[0].endDate]

    # sim plotting
    handles = []
    for l, c, ls, m, obj in zip(labels, colors, linestyles, markers, objects):
        simDF = evt.evaluator.Evaluator.get_sim(
            obj, forecastDay=forecastDay, start_end=start_end)[station_list]
        curves = _Curves(pd.to_datetime(simDF.index), ax=ax)
        for sta, y in simDF.iteritems():
            curves.add_curve(sta, y, {'color': c, 'linestyle': '-',
                                      'marker': m})
        if plot_type is 'median':
            curves.plot_median(style={'color': c, 'linestyle': '-',
                                      'marker': m},
                               envelope=envelope)
        elif plot_type is 'mean':
            curves.plot_median(style={'color': c, 'linestyle': '-',
                                      'marker': m},
                               envelope=envelope)
        else:
            curves.plot(legend=False)
        handles.append(mlines.Line2D([], [], color=c, label=l))
        if outputCSV is not None:
            curves.write_csv(outputCSV, model=l)

    plt.ylabel(ylabel)
    plt.title(title, loc='left', x=0.05)
    plt.tight_layout()

    # obs are plotted only for the first object in the list
    if type(objects[0]) is evt.evaluator.Evaluator:
        obsDF = objects[0].get_obs(forecastDay=forecastDay,
                                   start_end=start_end)[station_list]
        curves = _Curves(pd.to_datetime(obsDF.index), ax=ax)
        for sta, y in obsDF.iteritems():
            curves.add_curve(sta, y, obs_style)
        if plot_type is 'median':
            curves.plot_median(style=obs_style,
                               envelope=envelope)
        elif plot_type is 'mean':
            curves.plot_median(style=obs_style,
                               envelope=envelope)
        else:
            curves.plot(legend=False)
        handles.append(mlines.Line2D([], [], **obs_style))
        if outputCSV is not None:
            curves.write_csv(outputCSV, model=obs_style['label'])

    # legend
    plt.figlegend(handles=handles, fontsize='small', ncol=legend_ncol)

    # set y boundaries
    ax.set_ylim(bottom=ymin, top=ymax)

    # set xticks
    Xticks = pd.to_datetime(simDF.index)
    fig.autofmt_xdate(bottom=0.17, rotation=70, ha='right')
    if xticking == 'mondays':
        xt = np.unique(Xticks.date)
        xt = xt[np.array([d.weekday() == 0 for d in xt])]
        ax.set_xticks(xt)
        ax.set_xticklabels([d.strftime('%Y-%m-%d') for d in xt])
        # ax.set_xticklabels([d.strftime('%d/%m/%y') for d in xt])
    elif xticking == 'daily':
        xt = np.unique(Xticks.date)
        ax.set_xticks(np.array(xt))
        ax.set_xticklabels([d.strftime('%Y-%m-%d') for d in xt])
        # ax.set_xticklabels([d.strftime('%d/%m/%y') for d in xt])
    else:
        locator = dates.AutoDateLocator(minticks=2)
        ax.xaxis.set_major_locator(locator)
        hfmt = dates.AutoDateFormatter(locator)
        ax.xaxis.set_major_formatter(hfmt)

    plt.xlim(left=Xticks[0], right=Xticks[-1])
    plt.tick_params(axis='x', which='major', labelsize='small')

    _set_minor_ticks(*nb_of_minor_ticks)

    # give more space to y label
    box = ax.get_position()
    ax.set_position([box.x0+box.x0*0.1, box.y0, box.width, box.height])

    # draw grid
    ax.grid(which='both', ls='-', c='#B0B0B0')

    # try to control number of yticks
    plt.locator_params(axis='y', nbins=10, min_n_ticks=10)

    # paint in black y=0 and x=0
    if black_axes is True:
        _black_axes(ax)

    # save figure
    _save_figure(outputFile, file_formats)


def plot_diurnalCycle(
        objects, station_list=None, availability_ratio=0.75,
        colors=None, linestyles=None, markers=None, outputFile=None,
        labels=None, title="", xlabel='Forecast time (hour)',
        ylabel='concentration', file_formats=['png'], outputCSV=None,
        ymin=None, ymax=None, normalize=False, plot_type=None, envelope=False,
        black_axes=False, obs_style={'color': 'k', 'alpha': 0.5}):
    """ Plot time series of obs and sim at a chosen forecast day.

    On each plot, there is one line per station and per object.
    Observations are taken from the fisrt object of the list.

    Parameters
    ----------
    objects : list of evaltools.Evaluator objects
        Evaluator objects used for plotting.
    station_list : None or list of str
        List of stations to display. If None, all stations
        of the first element of <objects> argument are
        processed.
    colors : None or list of str
        Line colors corresponding to each objects.
    linestyles : None or list of str
        Line styles corresponding to each objects.
    markers : None or list of str
        Line markers corresponding to each objects.
    outputFile : str or None
        File where to save the plots (without extension). If None, the figure
        is shown in a popping window.
    labels : None or list of str
        List of labels for the legend.
    title : str
        Title for the plot.
    xlabel, ylabel : str
        Label for x and y axes.
    file_formats : list of str
        List of file extensions.
    outputCSV : str or None
        File where to save the data. The file name must contain {model}
        instead of the model name.
    ymin, ymax : None or scalar
        Limits of the axes.
    normalize : bool
        If True, values are normalized for each station by substracting the
        mean and dividing by the standard deviation of the diurnal cycle
        values for the station.
    plot_type : str
        If 'median' the median of all stations values at a forecast hour
        is plotted. If 'mean' the mean of all stations values at a forecast
        hour is plotted. For any other value, all station values are plotted
        separately.
    envelope : bool
        Only use if plot_type is 'median' or 'mean'. If True, draw quartiles
        one and three around the median curve (case plot_type == 'median')
        or draw +/- the standard deviation around the mean curve (case
        plot_type == 'mean').
    black_axes : bool
        If true, y=0 and x=0 lines are painted in black.
    obs_style : dict
        Dictionary of arguments passed to pyplot.plot for observation
        curve. Default value is {'color': 'k', 'alpha': 0.5}.

    """
    # check if objects are defined on the same period and same stations
    _checkPeriod(objects)
    _checkStationList(objects)

    # define default parameters
    labels, colors, linestyles, markers, legend_ncol, seriesType = \
        _defaultParams(objects, labels, colors, linestyles, markers)
    if station_list is None:
        station_list = reduce(np.intersect1d,
                              [obj.stations.index for obj in objects])
    os = dict(obs_style)
    if not 'label' in os:
        os['label'] = 'observations'

    # plotting
    plt.close('all')
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(1, 1, 1)

    # sim plotting
    handles = []
    for l, c, ls, m, obj in zip(labels, colors, linestyles, markers, objects):
        df = obj.FTscores(
            score_list=['sim_median'],
            availability_ratio=availability_ratio)[
                'sim_median'].loc[station_list]
        df.__delitem__('coord')
        df.columns = range(obj.forecastHorizon*24)
        if normalize is True:
            df = (df.subtract(df.mean(axis=1), axis=0)).divide(df.std(axis=1),
                                                               axis=0)
        if plot_type is 'median':
            df = df.quantile([0.25, 0.5, 0.75], axis=0).T
            plt.plot(df.index, df[0.5], color=c, linestyle='-', marker=m)
            if envelope is True:
                plt.plot(df.index, df[[0.25, 0.75]],
                         color=c, linestyle='--', marker=m)
        elif plot_type is 'mean':
            df_std = df.std(axis=0)
            df = df.mean(axis=0)
            plt.plot(df.index, df, color=c, linestyle='-')
            if envelope is True:
                plt.plot(df.index, df + df_std, color=c, linestyle='--')
                plt.plot(df.index, df - df_std, color=c, linestyle='--')
        else:
            plt.plot(df.columns, df.T, color=c, linestyle=ls, marker=m)
        handles.append(mlines.Line2D([], [], color=c, label=l))
        if outputCSV is not None:
            df.to_csv(outputCSV.format(model=l), sep=' ', na_rep='nan',
                      float_format='%g', header=True, index=True,
                      date_format='%Y%m%d%H')

    plt.ylabel(ylabel)
    plt.title(title, loc='left', x=0.05)
    plt.tight_layout()

    # obs are plotted only for the first object in the list
    df = obj.FTscores(
        score_list=['obs_median'],
        availability_ratio=availability_ratio)['obs_median'].loc[station_list]
    df.__delitem__('coord')
    X = range(obj.forecastHorizon*24)
    df.columns = X
    if normalize is True:
        df = (df.subtract(df.mean(axis=1), axis=0)).divide(df.std(axis=1),
                                                           axis=0)
    if plot_type is 'median':
        df = df.quantile([0.25, 0.5, 0.75], axis=0).T
        os.update({'linestyle': '-'})
        plt.plot(df.index, df[0.5], **os)
        if envelope is True:
            os.update({'linestyle': '--'})
            plt.plot(df.index, df[[0.25, 0.75]], **os)
    elif plot_type is 'mean':
        df_std = df.std(axis=0)
        df = df.mean(axis=0)
        os.update({'linestyle': '-'})
        plt.plot(df.index, df, **os)
        if envelope is True:
            os.update({'linestyle': '--'})
            plt.plot(df.index, df + df_std, **os)
            plt.plot(df.index, df - df_std, **os)
    else:
        plt.plot(df.columns, df.T, color='k', alpha=0.5)
    handles.append(mlines.Line2D([], [], **os))
    if outputCSV is not None:
        df.to_csv(outputCSV.format(model=os['label']), sep=' ',
                  na_rep='nan', float_format='%g', header=True, index=True,
                  date_format='%Y%m%d%H')

    # x label
    plt.xlabel(xlabel)

    # give more space to x and y labels
    box = ax.get_position()
    ax.set_position([box.x0+box.x0*0.2, box.y0+box.y0*0.2,
                     box.width, box.height])

    # legend
    plt.figlegend(handles=handles, fontsize='small', ncol=legend_ncol)

    # set y boundaries
    ax.set_ylim(bottom=ymin, top=ymax)

    # set xticks
    plt.xlim(left=X[0], right=X[-1])
    major_ticks = np.arange(0, len(X), 12)
    minor_ticks = np.arange(0, len(X), 3)
    ax.set_xticks(major_ticks)
    ax.set_xticks(minor_ticks, minor=True)

    # add one minor ytick between each major one
    minorLocator = AutoMinorLocator(n=2)
    ax.yaxis.set_minor_locator(minorLocator)

    # add three minor xtick between each major one
    minorLocator = AutoMinorLocator(n=4)
    ax.xaxis.set_minor_locator(minorLocator)

    # draw grid
    ax.grid(which='both', ls='-', c='#B0B0B0')

    # try to control number of yticks
    plt.locator_params(axis='y', nbins=10, min_n_ticks=10)

    # paint in black y=0 and x=0
    if black_axes is True:
        _black_axes(ax)

    # save figure
    _save_figure(outputFile, file_formats)


def plot_stationScoreDensity(
        objects, score, forecastDay=0, labels=None,
        colors=None, linestyles=None, outputFile=None, title="",
        file_formats=['png'], availability_ratio=0.75, nb_stations=False):
    """ Draw the probability density of a score computed per station.

    This function is based on kernel density estimation.

    Parameters
    ----------
    objects : list of evaltools.Evaluator objects
        Evaluator objects used for plotting.
    score : str
        Computed score.
    forecastDay : int.
        Integer corresponding to the chosen forecastDay used for plotting.
    labels : list of str
        List of labels for the legend.
    colors : None or list of str
        Line colors corresponding to each objects.
    linestyles : None or list of str
        Line styles corresponding to each objects.
    outputFile : str or None
        File where to save the plots (without extension). If None, the figure
        is shown in a popping window.
    title : str
        Title for the plots. It must contain {score} instead of the
        score name.
    file_formats : list of str
        List of file extensions.
    availability_ratio : float
        Minimal rate of data available on the period required per forecast
        term to compute the mean scores.
    nb_stations : bool
        If True, the number of stations used to estimate the density is
        displayed in the legend.

    """
    # check if objects are defined on the same period and same stations
    _checkPeriod(objects)
    _checkStationList(objects)

    # define default parameters
    markers = None
    labels, colors, linestyles, markers, legend_ncol, seriesType = \
        _defaultParams(objects, labels, colors, linestyles, markers)

    plt.close('all')
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(1, 1, 1)
    for obj, lab, c, ls in zip(objects, labels, colors, linestyles):
        # compute scores
        df = obj.stationScores(
            [score],
            availability_ratio=availability_ratio)['D{}'.format(forecastDay)]
        # plotting
        if nb_stations is True:
            df.columns = [lab + " ({} stations)".format(
                          sum(df[score].notna()))]
        else:
            df.columns = [lab]
        try:
            df.plot.kde(ax=ax, color=c, linestyle=ls)
        except (ValueError, np.linalg.linalg.LinAlgError):
            print(("Warning: KDE for {} can't be computed (scores " +
                   "may contain Inf values)").format(lab))

    ax.grid(which='both', ls='-', c='#B0B0B0')
    plt.title(title)

    # save figure
    _save_figure(outputFile, file_formats)


def plot_dataDensity(objects, forecastDay=0, labels=None, colors=None,
                     linestyles=None, outputFile=None, title="",
                     file_formats=['png'], xmin=None, xmax=None,
                     obs_style={'color': 'k', 'alpha': 0.5}):
    """ Probability density function.

    Draw the probability density of observed (for the first object only
    and simulated data. This function is based on kernel density estimation.

    Parameters
    ----------
    objects : list of evaltools.Evaluator objects
        Evaluator objects used for plotting.
    forecastDay : int.
        Integer corresponding to the chosen forecastDay used for plotting.
    labels : list of str
        List of labels for the legend.
    colors : None or list of str
        Line colors corresponding to each objects.
    linestyles : None or list of str
        Line styles corresponding to each objects.
    outputFile : str or None
        File where to save the plots (without extension). If None, the figure
        is shown in a popping window.
    title : str
        Title for the plots. It must contain {score} instead of the
        score name.
    file_formats : list of str
        List of file extensions.
    xmin, xmax : None or scalar
        Limits of the x axis.
    obs_style : dict
        Dictionary of arguments passed to pyplot.plot for observation
        curve. Default value is {'color': 'k', 'alpha': 0.5}.

    """
    # check if objects are defined on the same period and same stations
    _checkPeriod(objects)
    _checkStationList(objects)

    # define default parameters
    markers = None
    labels, colors, linestyles, markers, legend_ncol, seriesType = \
        _defaultParams(objects, labels, colors, linestyles, markers)
    if not 'label' in obs_style:
        obs_style['label'] = 'observations'

    plt.close('all')
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(1, 1, 1)
    for obj, lab, c, ls in zip(objects, labels, colors, linestyles):
        # compute scores
        sim = obj.get_sim(forecastDay=forecastDay)
        df = pd.DataFrame(sim.values.flatten())
        # plotting
        df.columns = [lab]
        try:
            df.plot.kde(ax=ax, color=c, linestyle=ls)
        except (ValueError, np.linalg.linalg.LinAlgError):
            print(("Warning: KDE for {} can't be computed (scores " +
                   "may contain Inf values)").format(lab))

    # obs are plotted only for the first object in the list
    obs = objects[0].get_obs(forecastDay=forecastDay)
    df = pd.DataFrame(obs.values.flatten())
    df.columns = [obs_style['label']]
    try:
        df.plot.kde(ax=ax, **obs_style)
    except (ValueError, np.linalg.linalg.LinAlgError):
        print("Warning: KDE for observations can't be computed " +
              "(time series may contain Inf values).")

    ax.grid(which='both', ls='-', c='#B0B0B0')
    plt.title(title)
    ax.set_xlim(left=xmin, right=xmax)

    # save figure
    _save_figure(outputFile, file_formats)
